#
# We can efficiently compute a (lower bound on) the size of a matrix group G
# by finding a vector v with an orbit on which G acts (faithfully). The
# following method does just that. The caller is responsible for supplying
# a suitable vector v (the shorter the orbit, the faster the computations).
#
SizeViaOrbit := function (G, v)
	local orbit, phi, size;

	orbit := Orbit(G, v);;
	orbit := ShallowCopy(orbit);; Sort(orbit);
	phi := ActionHomomorphism(G, orbit);
	size := Size(Image(phi));
	Unbind(phi); Unbind(orbit);
	return size;
end;;

#
# Find a nice permutation presentation of a matrix group by its
# action on the orbit of a given vector.
#
IsomorphismPermGroupViaOrbit := function (G, v)
	local orbit, phi;

	Print("  Computing orbit...\n");
	orbit := Orbit( G, v );;
	orbit := ShallowCopy(orbit);;
	Sort(orbit);
	#
	Print("  Computing permutation group ...\n");
	phi := ActionHomomorphism( G, orbit, "surjective" );;
	#
	Print("  Verifying that perm group is isomorphic image of the stabilizer...\n");
	Assert(0, IsSurjective(phi) and Size(Image(phi)) = Size(G));
	#
	return phi;
end;;

#
# Find relators of a finite presentation on FG_gens of the group H generated
# by H_gens, where each element of H_gens is mapped to the corresponding
# element of FG_gens.
#
GroupRelatorsViaOrbit := function( H, H_gens, FG_gens )
	local psi, Fgens, Frels;

	Print("  Determining finite presentation...\n");
	psi := IsomorphismFpGroupByGenerators( H, H_gens );;

	# Determine the relators
	Fgens := FreeGeneratorsOfFpGroup(Image(psi));
	Frels := RelatorsOfFpGroup(Image(psi));

	# Verify all relations actually hold in H
	Assert(0, ForAll( Frels, r->MappedWord( r, Fgens, H_gens ) = One(H) ));
	return List( Frels, r->MappedWord( r, Fgens, FG_gens ) );
end;;

#
# Auxillary function used by G2 and Sp6 below. Computes a list of
# generators for G2 in the representation we need.
#
G2_Sp_helper := function (q, TMP)
	local F, gens, g, g2, x, i, j, G;

	Assert(0, IsEvenInt(q));

	F := GF(q);
	gens := [];

	for g in GeneratorsOfGroup(TMP(3,F)) do
		g2 := TransposedMat(g)^-1;
		x := NullMat(6,6,F);
		for i in [1..3] do
			for j in [1..3] do
				x[i][j] := g[i][j];
				x[3+i][3+j] := g2[i][j];
			od;
		od;
		Add(gens, x);
	od;

	x := [
		[ 1,  0,  0,  1,  0,  0],
		[ 0,  1,  0,  0,  0,  0],
		[ 0,  0,  1,  0,  0,  0],
		[ 0,  0,  0,  1,  0,  0],
		[ 0,  0,  1,  0,  1,  0],
		[ 0, -1,  0,  0,  0,  1] ] * One(F);
	Add(gens, x);

	return Group(gens);
end;;

#
# Return a suitable representation of the group G2(q),
# where q must be a power of two.
#
G2 := function(q)
	local G;
	G := G2_Sp_helper(q, SL);
	#SetSize(G, q^6*(q^6-1)*(q^2-1));
	return G;
end;;

#
# Return a suitable representation of the group Sp(6,q),
# where q must be a power of two.
#
Sp6 := function(q)
	local G, x;
	G := G2_Sp_helper(q, GL);
	if q = 2 then
		x := [
			[ 1,  0,  0,  1,  0,  0],
			[ 0,  1,  0,  0,  1,  0],
			[ 0,  0,  1,  0,  0,  1],
			[ 0,  0,  0,  1,  0,  0],
			[ 0,  0,  0,  0,  1,  0],
			[ 0,  0,  0,  0,  0,  1] ] * One(GF(q));
		G := ClosureGroup(G, x);
	fi;
	#SetSize(G, q^6*(q^6-1)*(q^2-1));
	return G;
end;;
